# See bottom of file for license and copyright information

=begin TML

---+ package Foswiki::LoginManager::CasLogin

This is a login manager that you can specify in the security setup section of
[[%SCRIPTURL{"configure"}%][configure]]. It provides users with a
trip to the CAS server to enter usernames and passwords.

Subclass of Foswiki::LoginManager; see that class for documentation of the
methods of this class.

=cut

package Foswiki::LoginManager::CasLogin;

use strict;
use warnings;
use Assert;
use AuthCAS;
use URI;
use CGI::Carp;
#use Data::Dumper;

use Foswiki::LoginManager ();
our @ISA = ('Foswiki::LoginManager');

=begin TML

---++ ClassMethod new ($session, $impl)

Construct the CasLogin object

=cut

sub new {
    my ( $class, $session ) = @_;
    my $this = $class->SUPER::new($session);
    $session->enterContext('can_login');
    if ( $Foswiki::cfg{Sessions}{ExpireCookiesAfter} ) {
        $session->enterContext('can_remember_login');
    }
    if ( $Foswiki::cfg{TemplateLogin}{PreventBrowserRememberingPassword} ) {
        $session->enterContext('no_auto_complete_login');
    }
    return $this;
}

# Pack key request parameters into a single value
# Used for passing meta-information about the request
# through a URL (without requiring passthrough)
sub _packRequest {
    my ( $uri, $method, $action ) = @_;
    return '' unless $uri;
    if ( ref($uri) ) {    # first parameter is a $session
        my $r = $uri->{request};
        $uri    = $r->uri();
        $method = $r->method() || 'UNDEFINED';
        $action = $r->action();
    }
    return "$method,$action,$uri";
}

# Unpack single value to key request parameters
sub _unpackRequest {
    my $packed = shift || '';
    my ( $method, $action, $uri ) = split( ',', $packed, 3 );
    return ( $uri, $method, $action );
}

=begin TML

---++ ObjectMethod forceAuthentication () -> $boolean

method called when authentication is required - redirects to (...|view)auth
Triggered on auth fail

=cut

sub forceAuthentication {
    my $this    = shift;
    my $session = $this->{session};

    unless ( $session->inContext('authenticated') ) {
        my $query = $session->{request};

        # Redirect with passthrough so we don't lose the original query params

        my $url = $session->getScriptUrl( 0, 'login' );

        # We use the query here to ensure the original path_info
        # from the request gets through to the login form. See also
        # PATH_INFO below.
        $url .= Foswiki::urlEncode( $query->path_info() );

        $query->param(
            -name  => 'foswiki_origin',
            -value => _packRequest($session)
        );
        $session->redirect( $url, 1 );    # with passthrough
        return 1;
    }
    return 0;
}

=begin TML

---++ ObjectMethod loginUrl () -> $loginUrl

Overrides LoginManager. Content of a login link.

=cut

sub loginUrl {
    my $this    = shift;
    my $session = $this->{session};
    my $topic   = $session->{topicName};
    my $web     = $session->{webName};
    return $session->getScriptUrl( 0, 'login', $web, $topic,
        foswiki_origin => _packRequest($session) );
}

=begin TML

---++ ObjectMethod login( $query, $session )

Redirect to CAS server to login. Successful login results in a username.

=cut

sub login {
    my ( $this, $query, $session ) = @_;
    my $users = $session->{users};

#my $d = Data::Dumper->new([\$query],['query']);
#   $d->Indent(1);
#carp( $d->Dump );
    my $origin = $query->param('foswiki_origin');
    my ( $origurl, $origmethod, $origaction ) = _unpackRequest($origin);
#carp ( "origurl 1: $origurl" );
#    my $loginName = $query->param('username');
    my $loginName;
#    my $loginPass = $query->param('password');
    my $remember  = $query->param('remember');
    my $casUrl = $Foswiki::cfg{CAS}{casUrl};
    my $CAFile = $Foswiki::cfg{CAS}{CAFile};
#carp( "casUrl: $casUrl" );
#carp( "CAFile: $CAFile" );
    my $cas = new AuthCAS(casUrl => $casUrl,
                          CAFile => $CAFile
                         );

    my $ughu = $origurl;
#carp( "ughu: $ughu" );
    my $ugh = URI->new($ughu);
    my $ughq = $ugh->query();
    $ughq = '' unless $ughq;
#carp( "ughq: $ughq" );
    my $ticket;
#    ($ticket = $ughu) =~ /ticket=(.*cas)/;
    if ($ughq =~ /ticket/) {
     my ($key,$value) = split('=',$ughq);
     $ticket = $value if $key eq 'ticket';
     $origurl = $ugh->path();
    }
     
    my $app_url = $Foswiki::cfg{DefaultUrlHost};
#carp("app_url: $app_url");
#carp("origurl 2: $origurl");

    # Eat these so there's no risk of accidental passthrough
#    $query->delete( 'foswiki_origin', 'username', 'password' );
    $query->delete( 'foswiki_origin', 'ticket' );

    # UserMappings can over-ride where the login template is defined
#    my $loginTemplate = $users->loginTemplateName();    #defaults to login.tmpl
#    my $tmpl =
#      $session->templates->readTemplate( $loginTemplate );

#    my $banner = $session->templates->expandTemplate('LOG_IN_BANNER');
    my $banner;
    my $note   = '';
    my $topic  = $session->{topicName};
    my $web    = $session->{webName};

    my $cgisession = $this->{_cgisession};

#    $cgisession->param( 'REMEMBER', $remember ) if $cgisession;
#    if (   $cgisession
#        && $cgisession->param('AUTHUSER')
#        && $loginName
#        && $loginName ne $cgisession->param('AUTHUSER') )
#    {
#        $banner = $session->templates->expandTemplate('LOGGED_IN_BANNER');
#        $note   = $session->templates->expandTemplate('NEW_USER_NOTE');
#    }

    my $error = '';

    if ($ticket) {
#        my $validation = $users->checkPassword( $loginName, $loginPass );
        my $validation = 1;
#        $error = $users->passwordError();

        $loginName = $cas->validateST($app_url.$origurl, $ticket);

carp( "loginName: $loginName  ticket: $ticket" );
        if ($loginName) {

            # SUCCESS our user is authenticated. Note that we may already
            # have been logged in by the userLoggedIn call in loadSession,
            # becuase the username-password URL params are the same as
            # the params passed to this script, and they will be used
            # in loadSession if no other user info is available.
            $this->userLoggedIn($loginName);
            $session->logEvent(
                'login',
                $web . '.' . $topic,
                "AUTHENTICATION SUCCESS - $loginName - "
            );

            # remove the sudo param - its only to tell TemplateLogin
            # that we're using BaseMapper..
            $query->delete('sudo');

            $cgisession->param( 'VALIDATION', $validation ) if $cgisession;
            if ( !$origurl || $origurl eq $query->url() ) {
                $origurl = $session->getScriptUrl( 0, 'view', $web, $topic );
            }
            else {

                # Unpack params encoded in the origurl and restore them
                # to the query. If they were left in the query string they
                # would be lost if we redirect with passthrough.
                if ( $origurl =~ s/\?(.*)// ) {
                    foreach my $pair ( split( /[&;]/, $1 ) ) {
                        if ( $pair =~ /(.*?)=(.*)/ ) {
                            $query->param( $1, TAINT($2) );
                        }
                    }
                }

                # Restore the action too
                $query->action($origaction) if $origaction;
            }

            # Restore the method used on origUrl so if it was a GET, we
            # get another GET.
            $query->method($origmethod);
            $session->redirect( $origurl, 1 );
            return;
        }
        else {
     ###
     ### Redirect the User for login at CAS server
     ###
     my $login_url = $cas->getServerLoginURL($app_url.$origurl);
#carp( "login_url: $login_url" );
     printf "Location: $login_url\n\n";
     exit 0;


#            # Tasks:Item1029  After much discussion, the 403 code is not
#            # used for authentication failures. RFC states: "Authorization
#            # will not help and the request SHOULD NOT be repeated" which
#            # is not the situation here.
#            $session->{response}->status(200);
#            $session->logEvent(
#                'login',
#                $web . '.' . $topic,
#                "AUTHENTICATION FAILURE - $loginName - "
#            );
#            $banner = $session->templates->expandTemplate('UNRECOGNISED_USER');
        }
    }
    else {

     ###
     ### Redirect the User for login at CAS server
     ###
     my $login_url = $cas->getServerLoginURL($app_url.$origurl);
#carp( "six: $login_url" );
     printf "Location: $login_url\n\n";
     exit 0;

        # If the loginName is unset, then the request was likely a perfectly
        # valid GET call to http://foswiki/bin/login
        # 4xx cannot be a correct status, as we want the user to retry the
        # same URL with a different login/password
#        $session->{response}->status(200);
    }

    # Remove the validation_key from the *passed through* params. It isn't
    # required, because the form will have a new validation key, and
    # giving the parameter twice will confuse the strikeone Javascript.
    $session->{request}->delete('validation_key');

    # set the usernamestep value so it can be re-displayed if we are here due
    # to a failed authentication attempt.
    $query->param( -name => 'usernamestep', -value => $loginName );

    # TODO: add JavaScript password encryption in the template
    $origurl ||= '';

    # Set session preferences that will be expanded when the login
    # template is instantiated
    $session->{prefs}->setSessionPreferences(
        FOSWIKI_ORIGIN => Foswiki::entityEncode(
            _packRequest( $origurl, $origmethod, $origaction )
        ),

        # Path to be used in the login form action.
        # Could have used %ENV{PATH_INFO} (after extending {AccessibleENV})
        # but decided against it as the path_info might have been rewritten
        # from the original env var.
        PATH_INFO => $query->path_info(),
        BANNER    => $banner,
        NOTE      => $note,
        ERROR     => $error
    );

#    my $topicObject = Foswiki::Meta->new( $session, $web, $topic );
#    $tmpl = $topicObject->expandMacros($tmpl);
#    $tmpl = $topicObject->renderTML($tmpl);
#    $tmpl =~ s/<nop>//g;
#    $session->writeCompletePage($tmpl);
}

1;
__DATA__
Module of Foswiki - The Free and Open Source Wiki, http://foswiki.org/

Copyright (C) 2008-2010 Foswiki Contributors. All Rights Reserved.
Foswiki Contributors are listed in the AUTHORS file in the root
of this distribution. NOTE: Please extend that file, not this notice.

Additional copyrights apply to some or all of the code in this
file as follows:

Copyright (C) 2005-2006 TWiki Contributors. All Rights Reserved.
Copyright (C) 2005 Greg Abbas, twiki@abbas.org

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version. For
more details read LICENSE in the root of this distribution.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

As per the GPL, removal of this notice is prohibited.
